#include <stdlib.h>

#include "pqueue.h"

struct pqueue
   {
   void ** data;
   size_t max;
   size_t size;
   bool (*comp)(void const * const a, void const * const b);
   };

////////////////////////////////////////////////////////////////////////////////
// comp should return true if *a comp *b.
// So, for a min priority queue, comp will be defined to return true if *a < *b,
// whereas for a max priority queue comp will be defined to return true if *a > *b,

struct pqueue * 
pqueue_new(size_t max, bool (*comp)(void const * const a, void const * const b))
   {
   if ( (0 == max) || (NULL == comp) )
      return NULL;

   struct pqueue * pq = malloc(sizeof(struct pqueue));
   if (NULL == pq)
      return NULL;

   pq->max = max;
   pq->size = 0;
   pq->comp = comp;

   // Allocate space for max+1 since we are going to waste the first element in order to use a one based array.
   pq->data = malloc((1 + pq->max) * sizeof(void *));
   if (NULL == pq->data)
      {
      free(pq);
      return NULL;
      }

   return pq;
   }


////////////////////////////////////////////////////////////////////////////////
void 
pqueue_delete(struct pqueue * pq)
   {
   if (NULL != pq)
      {
      free(pq->data);
      free(pq);
      }
   }

////////////////////////////////////////////////////////////////////////////////
void * 
pqueue_push(struct pqueue * pq, void * elem)
   {
   if ( (NULL == pq) || (NULL == elem) )
      return NULL;

   if (pq->size == pq->max)
      return NULL;

   pq->size++;
   pq->data[pq->size] = elem;

   // sift up
   void * temp;
   size_t i, parent;
   for (i = pq->size, parent = i/2;
	(i > 1) && (! pq->comp(pq->data[parent], pq->data[i]));
	i = parent, parent = i/2)
      {
      temp = pq->data[i];
      pq->data[i] = pq->data[parent];
      pq->data[parent] = temp;
      }

   return elem;
   }


////////////////////////////////////////////////////////////////////////////////
void * 
pqueue_pop(struct pqueue * pq)
   {
   if ( (NULL == pq) || (0 == pq->size) )
      return NULL;
   
   void * val = pq->data[1];

   pq->data[1] = pq->data[pq->size];
   pq->size--;

   // sift down
   void * temp;
   size_t i, child;
   for (i = 1;  (child = 2*i) <= pq->size; i = child)
      {
      if ( (child+1 <= pq->size) && (pq->comp(pq->data[child+1], pq->data[child])) )
	   child++;
      if (pq->comp(pq->data[child], pq->data[i]))
	 {
	 temp = pq->data[i];
	 pq->data[i] = pq->data[child];
	 pq->data[child] = temp;
	 }
      else
	 break;
      }

   return val;
   }


////////////////////////////////////////////////////////////////////////////////
void *
pqueue_peek(struct pqueue * pq)
   {
   if ( (NULL == pq) || (0 == pq->size) )
      return NULL;

   return pq->data[1];
   }

////////////////////////////////////////////////////////////////////////////////
size_t
pqueue_size(struct pqueue * pq)
   {
   if (NULL == pq)
      return 0;

   return  pq->size;
   }
